package nodeinfo

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"
)

type fetcher struct {
	domain    string
	client    *http.Client
	matrix    bool
	nodeinfo  bool
	nodeinfoX bool
}

func newFetcher(domain string) *fetcher {
	return &fetcher{
		domain: domain,
		// don't follow redirects
		client: &http.Client{
			CheckRedirect: func(req *http.Request, via []*http.Request) error {
				return http.ErrUseLastResponse
			},
		},
		matrix:    true,
		nodeinfo:  true,
		nodeinfoX: true,
	}
}

func (f *fetcher) validate() error {
	// validate options
	if !(f.nodeinfo || f.matrix) {
		return fmt.Errorf("all supported discovery protocols are disabled")
	}

	// validate user input
	rawDomain := f.domain
	if !(strings.HasPrefix(rawDomain, "http://") || strings.HasPrefix(rawDomain, "https://")) {
		rawDomain = "https://" + rawDomain
	}
	domain, err := url.Parse(rawDomain)
	if err != nil {
		return fmt.Errorf("parsing domain failed: %w", err)
	}
	f.domain = domain.Host

	return nil
}

func (f *fetcher) do() (*Node, error) {
	var el []error

	if f.nodeinfo {
		n, found, err := f.doNodeInfo()
		if found {
			return n, err
		}
		el = append(el, err)
	}

	if f.nodeinfoX {
		n, found, err := f.doNodeInfoX()
		if found {
			return n, err
		}
		el = append(el, err)
	}

	if f.matrix {
		n, found, err := f.doMatrix()
		if found {
			return n, err
		}
		el = append(el, err)
	}

	return nil, fmt.Errorf("could not detect well-known service: %w", errors.Join(el...))
}

type nodeInfoLookup struct {
	Links []struct {
		Schema string `json:"rel"`
		Href   string `json:"href"`
	} `json:"links"`
}

func (f *fetcher) doNodeInfo() (*Node, bool, error) {
	// A client should first try the HTTPS protocol and fall back to HTTP on connection errors or if it can't validate the presented certificate.
	// TODO: check if https or http
	schema := "https"

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/nodeinfo", schema, f.domain))
	if err != nil {
		return nil, false, err
	}

	// no nodeinfo server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get nodeinfo endpoints
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	lookup := nodeInfoLookup{}
	if err := json.Unmarshal(rawResp, &lookup); err != nil {
		return nil, true, err
	}

	links := make(map[string]string)
	for _, link := range lookup.Links {
		links[link.Schema] = link.Href
	}
	if len(links) == 0 {
		return nil, true, fmt.Errorf("broken lookup endpoint, no links found")
	}

	var href string
	if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.1"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/2.0"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/1.1"]; ok {
		href = ref
	} else if ref, ok := links["http://nodeinfo.diaspora.software/ns/schema/1.0"]; ok {
		href = ref
	} else {
		return nil, true, fmt.Errorf("broken lookup endpoint, unknown schema(s) found: %s", links)
	}

	resp, err = f.client.Get(href)
	if err != nil {
		return nil, false, err
	}
	rawResp, err = io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}

	node := &Node{}
	if err := json.Unmarshal(rawResp, node); err != nil {
		return nil, true, err
	}
	return node, true, nil
}

type xNodeInfo struct {
	Server struct {
		Software string `json:"software"`
		Version  string `json:"version"`
		Name     string `json:"name"`
		BaseURL  string `json:"baseUrl"`
	} `json:"server"`
	Version           string   `json:"version"`
	OpenRegistrations bool     `json:"openRegistrations"`
	Protocols         []string `json:"protocols"`
	Services          Services `json:"services"`
	Usage             struct {
		Users struct {
			Total          int64 `json:"total"`
			ActiveHalfyear int64 `json:"activeHalfyear"`
			ActiveMonth    int64 `json:"activeMonth"`
			ActiveWeek     int64 `json:"activeWeek"`
		} `json:"users"`
	} `json:"usage"`
}

func (f *fetcher) doNodeInfoX() (*Node, bool, error) {
	// A client should first try the HTTPS protocol and fall back to HTTP on connection errors or if it can't validate the presented certificate.
	// TODO: check if https or http
	schema := "https"

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/x-nodeinfo2", schema, f.domain))
	if err != nil {
		return nil, false, err
	}

	// no x-nodeinfo server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get nodeinfo endpoints
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	node := &xNodeInfo{}
	if err := json.Unmarshal(rawResp, node); err != nil {
		return nil, true, err
	}

	infoNode := &Node{
		Version: "2.1",
		Software: Software{
			Name:    node.Server.Software,
			Version: node.Server.Version,
		},
		OpenRegistrations: node.OpenRegistrations,
		Protocols:         node.Protocols,
		Services:          node.Services,
		Usage: Usage{
			Users: UsersUsage{
				Total:          node.Usage.Users.Total,
				ActiveHalfyear: node.Usage.Users.ActiveHalfyear,
				ActiveMonth:    node.Usage.Users.ActiveMonth,
			},
		},
		Metadata: map[string]interface{}{
			"serverName":           node.Server.Name,
			"baseUrl":              node.Server.BaseURL,
			"UsageUsersActiveWeek": node.Usage.Users.ActiveWeek,
		},
	}

	return infoNode, true, nil
}

type matrixLookup struct {
	ServerAddr string `json:"m.server"`
}

type matrixVersion struct {
	Server struct {
		ImplName    string `json:"name"`
		ImplVersion string `json:"version"`
	} `json:"server"`
}

func (f *fetcher) doMatrix() (*Node, bool, error) {
	schema := "https"
	node := Node{
		Protocols: []string{ProtocolMatrix},
	}

	// TODO: respect DNS lookup as per https://spec.matrix.org/v1.8/server-server-api/#resolving-server-names

	resp, err := f.client.Get(fmt.Sprintf("%s://%s/.well-known/matrix/server", schema, f.domain))
	if err != nil {
		return nil, false, err
	}

	// no matrix server found
	if resp.StatusCode == http.StatusNotFound {
		return nil, false, nil
	} else if resp.StatusCode != http.StatusOK {
		return nil, false, fmt.Errorf("unexpected status code on lookup")
	}

	// get matrix server address
	rawResp, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	lookup := matrixLookup{}
	if err := json.Unmarshal(rawResp, &lookup); err != nil {
		return nil, true, err
	}

	// get server software
	resp, err = f.client.Get(fmt.Sprintf("%s://%s/_matrix/federation/v1/version", schema, lookup.ServerAddr))
	if err != nil {
		return nil, false, err
	}
	rawResp, err = io.ReadAll(resp.Body)
	if err != nil {
		return nil, true, err
	}
	version := matrixVersion{}
	if err := json.Unmarshal(rawResp, &version); err != nil {
		return nil, true, err
	}
	node.Software.Version = version.Server.ImplVersion
	node.Software.Name = version.Server.ImplName
	if soft, found := mapMatrixSoftware[strings.ToLower(version.Server.ImplName)]; found {
		node.Software.Homepage = soft.www
		node.Software.Repository = soft.git
	}

	node.OpenRegistrations = false // TODO

	return &node, true, nil
}
