package nodeinfo

import (
	"fmt"
	"net/http"
)

// Node metadata about a server running in the federation
type Node struct {
	Software          Software               `json:"software"`
	Version           string                 `json:"version"`
	OpenRegistrations bool                   `json:"openRegistrations"`
	Protocols         []string               `json:"protocols"`
	Services          Services               `json:"services"`
	Usage             Usage                  `json:"usage"`
	Metadata          map[string]interface{} `json:"metadata"`
}

// Software is about node software
type Software struct {
	Name       string `json:"name"`
	Version    string `json:"version"`
	Repository string `json:"repository"`
	Homepage   string `json:"homepage"`
}

// Services tell about third party sites the node can connect to or interact with
type Services struct {
	Inbound  []string `json:"inbound"`
	Outbound []string `json:"outbound"`
}

// Usage are usage statistics for the node
type Usage struct {
	Users         UsersUsage `json:"users"`
	LocalPosts    int64      `json:"localPosts,omitempty"`
	LocalComments int64      `json:"localComments,omitempty"`
}

// UsersUsage are statistics about the users of the node
type UsersUsage struct {
	Total          int64 `json:"total,omitempty"`
	ActiveHalfyear int64 `json:"activeHalfyear,omitempty"`
	ActiveMonth    int64 `json:"activeMonth,omitempty"`
}

// QueryNode fetch node info based on domain
func QueryNode(domain string, opts ...QueryNodeOption) (*Node, error) {
	f := newFetcher(domain)

	for _, fn := range opts {
		if err := fn(f); err != nil {
			return nil, fmt.Errorf("could not apply option: %w", err)
		}
	}

	if err := f.validate(); err != nil {
		return nil, fmt.Errorf("validation of fetcher failed: %w", err)
	}

	return f.do()
}

// QueryNodeOption are functions to be used as options for QueryNode
type QueryNodeOption func(f *fetcher) error

// QueryOptionClient return an option for QueryNode that can set an specific http client
func QueryOptionClient(c *http.Client) QueryNodeOption {
	return func(f *fetcher) error {
		f.client = c
		return nil
	}
}

// QueryOptionNoNodeinfo do not discovery based on the nodeinfo2 specification
func QueryOptionNoNodeinfo() QueryNodeOption {
	return func(f *fetcher) error {
		f.nodeinfo = false
		f.nodeinfoX = false
		return nil
	}
}

// QueryOptionNoMatrix do not discovery based on the matrix specification
func QueryOptionNoMatrix() QueryNodeOption {
	return func(f *fetcher) error {
		f.matrix = false
		return nil
	}
}
